#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sstream>
#include <vector>

#include "types.h"
extern "C" {
#include "state.h"
}
#include "fceu.h"

#define BISQBOT_INTERNAL
#include "bisqbotdefs.hh"
#include "bisqbot.hh"

#include "coroutine.h"

int BotFrontDisableVideo=0;

struct SaveState
{
    std::vector<unsigned char> Data;
    const BisqBotStateBase* Bot;
    
    SaveState() : Bot(0) { }
    ~SaveState() { delete Bot; }

    void Create()
    {
        delete Bot;
        
        FILE*fp = tmpfile();
        FCEUSS_SaveFP(fp);
        rewind(fp);
        
        fseek(fp,0,SEEK_END);
        long pos = ftell(fp);
        rewind(fp);
        
        SaveState result;
        Data.resize(pos);
        fread(&Data[0], 1, pos, fp);
        fclose(fp);
        
        Bot = BisqBotNewSaveState();
    }
    void Load()
    {
        FILE*fp = tmpfile();
        fwrite(&Data[0], 1, Data.size(), fp);
        rewind(fp);
        FCEUSS_LoadFP(fp, 0);
        fclose(fp);
        
        BisqBotLoadState(Bot);
    }
private:
    SaveState(const SaveState& b);
    void operator=(const SaveState& b);
};

static unsigned char CurInput;
static bool Active=false;

static void DisplaySeq(const std::vector<unsigned char>& Seq)
{
    std::stringstream msg;
    for(unsigned a=0; a<Seq.size(); )
    {
        unsigned c=1, ch=Seq[a];
        while(a+c < Seq.size() && Seq[a+c] == ch) ++c;
        msg << c; msg << '*';
        
        unsigned b=0;
        if(ch & K_A) { if(b++) msg << ("_"); msg << ("A"); }
        if(ch & K_B) { if(b++) msg << ("_"); msg << ("B"); }
        if(ch & K_SE) { if(b++) msg << ("_"); msg << ("SE"); }
        if(ch & K_ST) { if(b++) msg << ("_"); msg << ("ST"); }
        if(ch & K_U) { if(b++) msg << ("_"); msg << ("U"); }
        if(ch & K_D) { if(b++) msg << ("_"); msg << ("D"); }
        if(ch & K_L) { if(b++) msg << ("_"); msg << ("L"); }
        if(ch & K_R) { if(b++) msg << ("_"); msg << ("R"); }
        if(!b) msg << ("00");
        msg << (", ");
        a += c;
    }
    BotFrontMsg("%s", msg.str().c_str());
}

#include <sys/file.h> /* For file locking */
void BotFrontMsg(const char*fmt, ...)
{
    FILE*fp;
    
    va_list ap;
    va_start(ap,fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    
    fp = fopen("rockman.log", "at");
    if(!fp) fp = fopen("rockman.log", "wt");
    if(!fp) return;
    flock(fileno(fp), LOCK_EX);
    fseek(fp,0,SEEK_END);

    va_start(ap,fmt);
    vfprintf(fp, fmt, ap);
    va_end(ap);
    
    fclose(fp);     
}

void BotFrontStart()
{
    fprintf(stderr,"BotFrontStart\n");
    BotFrontDisableVideo=1;
    Active=true;
}

void BotFrontEnd()
{
    fprintf(stderr,"BotFrontEnd\n");
    BisqBotReset();
    BotFrontDisableVideo=0;
    Active=false;
} 

void BotFrontToggle()
{
    if(!Active)
        BotFrontStart();
    else
        BotFrontEnd();
}

int BotFrontActive()
{
    return Active ? 1 : 0;
}

unsigned BotFrontGetKeymask()
{
    if(Active)
    {
        //fprintf(stderr, "CurInput: %02X\n", CurInput);
        return CurInput;
    }
    //fprintf(stderr, "Nul keymask\n");
    return 0;
}

/*
    FCEU main loop:
        1. Emulate frame     
        2. Read keyboard input (BotFrontIter)
        3. Update joypad input (BotFrontGetKeymask)

    It is actually possible for keyboard input to be
    read multiple times per frame... Currently we hope
    this don't happen.
*/

struct Attempt
{
    std::vector<unsigned char> Seq;
    SaveState anchor;
    unsigned retry_count;
    unsigned retry_max;
    unsigned frame_begin;
    unsigned frame_count;
    unsigned frame_max;
    
    void Init(unsigned Step, unsigned Frameno)
    {
        Seq.resize(Frameno);
        anchor.Create();
        retry_count = 0;
        retry_max   = MAX_REFINES[Step];
        frame_begin = Frameno;
        frame_count = Frameno;
        frame_max   = MAX_LENGTH[Step];
    }
    void Reset()
    {
        Seq.resize(frame_begin);
        frame_count = frame_begin;
        anchor.Load();
        if(frame_count == 0) BisqBotReset();
        ++retry_count;
    }
};

void BotFrontIter()
{
    scrBegin;
Restart:
    while(!Active)
    {
        BisqBotEvaluate(0);
        BisqBotReset();
        scrReturnV;
    }

    BotFrontMsg("--------------- New run  ------------------ \n");

    #define RunFrameWith(k) \
        CurInput = (k); \
        FCEUI_FrameAdvance(); \
        scrReturnV; if(!Active)goto Restart;
    
    static std::vector<unsigned char> BestSeq;
    static unsigned long long BestScore;
    
    static Attempt Steps[MAX_PARTS];
    static unsigned CurStep;
    
    if(!FCEUI_EmulationPaused()) FCEUI_ToggleEmulationPause();
    
    CurStep=0;
    Steps[0].Init(0, 0);
    Steps[0].Reset();

    BestSeq.clear();
    BestScore = 0;
    
    /* MAIN LOOP */
    for(;;)
    {
        /* All frames handled? */
        if(Steps[CurStep].frame_count >= Steps[CurStep].frame_max)
        {
        TryOver:
            /* Return to the beginning of this step (and possibly,
             * to the end of the previous step)
             */
            Steps[CurStep].Reset();

            /* Check retry counter */
            if(Steps[CurStep].retry_count < Steps[CurStep].retry_max)
            {
                /* One more time */
                ++Steps[CurStep].retry_count;
                continue;
            }
            /* All retries exhausted */
            if(CurStep > 0)
            {
                BotFrontMsg("*** Return step %u->%u\n", CurStep,CurStep-1);
                /* Back to previous level */
                --CurStep;
                goto TryOver;
            }
            /* End loop */
            break;
        }
        
        /* Run frame. Scope the "k" off. */
        if(true)
        {
            static unsigned char k;
            //fprintf(stderr, "Frame %u\r", Steps[CurStep].frame_count);
            k = BisqBotInput(Steps[CurStep].frame_count++);
            Steps[CurStep].Seq.push_back(k);
            RunFrameWith(k);
        }
        
        /* Let BisqBot analyze how we're doing */
        unsigned long long Score = BisqBotEvaluate(Steps[CurStep].frame_count);
        bool Restart=false;
        switch(BisqBotTick(Steps[CurStep].frame_count, CurStep))
        {
            case BisqBotRestart:
                CurStep=0; // passthru
            case BisqBotRestartStep:
            {
                /* Try over */
                Restart=true;
                break;
            }
            case BisqBotContinue:
            {
                /* ok */
                break;
            }
            case BisqBotNextStep:
            {
                /* Go to next step */
                if(CurStep+1 < MAX_PARTS)
                {
                    BotFrontMsg("*** Advance step %u->%u at frame %u (score %llu)\n",
                        CurStep,CurStep+1, Steps[CurStep].frame_count, Score);
                    DisplaySeq(Steps[CurStep].Seq);
                    BotFrontMsg("\n");

                    Steps[CurStep+1].Init(CurStep+1, Steps[CurStep].frame_count);
                    Steps[CurStep+1].Seq = Steps[CurStep].Seq;
                    ++CurStep;
                }
                break;
            }
            case BisqBotAccept:
            {
                /* Bingo */
                const char* beep = "";
                if(Score >= BestScore) beep = "";
                
                BotFrontMsg("\n*** %s Bingo in %u frames (score: %llu):\n",
                    beep, Steps[CurStep].frame_count, Score);
                DisplaySeq(Steps[CurStep].Seq);
                BotFrontMsg("\n\n");
                
                Restart=true;
                break;
            }
        }
        if(Score > BestScore)
        {
            BestScore = Score;
            BestSeq   = Steps[CurStep].Seq;
        }
        if(Restart) goto TryOver;
    }
    
    /* Lastly, apply the best sequence. Shield with a scope again. */
    if(true)
    {
        Steps[0].Reset();
        
        static unsigned frameno;
        for(frameno = 0; frameno < BestSeq.size(); ++frameno)
        {
            RunFrameWith(BestSeq[frameno]);
        }
    }
    
    #undef RunFrameWith

End: ;
    BotFrontEnd();

    scrFinishV;
}
