#ifndef bqtSndTool2MIDIhh
#define bqtSndTool2MIDIhh

#include <vector>
#include <string>

/*

MIDI events:

    8c note volume    KEY OFF
    9c note volume    KEY ON
    Ac note volume    KEY AFTERTOUCH
    Bc ctrl value     CONTROL
    Cc patch          PATCH CHANGE
    Dc ??             CHANNEL AFTERTOUCH
    Ec lo7 hi7        PITCH WHEEL

*/

class MIDItrack
{
protected:
    unsigned delay;
    unsigned char running;
    unsigned readpos;
    
    std::vector<unsigned char> track;
public:
    MIDItrack() : delay(0), running(0), readpos(0)
    {
    }
    void Reset() { delay=0; running=0; readpos=0; track.clear(); }
    
    void Add(unsigned char data);
    void Flush();
    void AddVarLen(unsigned t);
    void AddText(unsigned char n, const char* t);
    void AddMarker(const char* s) { AddText(6, s); }
    void AddCopyright(const char* s) { AddText(2, s); }
    void AddFilename(const char* s) { AddText(1, s); }
    void AddSongname(const char* s) { AddText(3, s); }
    void AddDelay(unsigned length) { delay += length; }
    void AddData(unsigned char data);
    void AddMeta(unsigned char metatype, unsigned char nbytes, ...);
    void AddData(unsigned char nbytes, unsigned char data, ...);
    
    void DecDelay(unsigned length) { delay -= length; }
    void ReadDelay();
    void ReadEventInto(class MIDIfile&, unsigned trackno);
    unsigned ReadVarLen();
    int ReadMeta(std::string& data);
    void PeekNextCmd(std::string& cmd, bool peek_only=true);
    
    unsigned size() const { return track.size(); }
    unsigned char operator[] (unsigned ind) const { return track[ind]; }
    const unsigned char* getdata() const { return (const unsigned char* )&track[0]; }
    
    void SetData(const unsigned char* buf, unsigned size)
    {
        Reset();
        track.assign(buf, buf+size);
    }
    
    unsigned GetDelay() const { return delay; }
};

class MIDIfile
{
protected:
    std::vector<MIDItrack> tracks;
    unsigned deltaticks, tempo;
    
    std::vector<unsigned char> midi;
    
    friend class MIDItrack;
public:
    MIDIfile() : deltaticks(1000), tempo(1000000)
    {
    }
    void AddLoopStart() { AddMarker(0, "loopStart"); }
    void AddLoopEnd() { AddMarker(0, "loopEnd"); }
    void AddText(unsigned trackno, unsigned char type, const char* s)
    {
        if(tracks.size() > trackno) tracks[trackno].AddText(type, s);
    }
    void AddMarker(unsigned trackno, const char* s) { AddText(trackno, 6, s); }
    void AddCopyright(unsigned trackno, const char* s) { AddText(trackno, 2, s); }
    void AddFilename(unsigned trackno, const char* s) { AddText(trackno, 1, s); }
    void AddSongname(unsigned trackno, const char* s) { AddText(trackno, 3, s); }
    void InitTrack(unsigned trackno);
    void Done();
    
    // Note: pressure != volume. Use ctrl(7) to set volume.
    void KeyOn(unsigned trackno, unsigned ch, unsigned char note, unsigned char pressure);
    
    void KeyOff(unsigned trackno, unsigned ch, unsigned char note, unsigned char pressure);

    // Nobody ever uses this. Use ctrl(7) to change the pressure.
    void KeyTouch(unsigned trackno, unsigned ch, unsigned char note, unsigned char pressure);

    void Control(unsigned trackno, unsigned ch, unsigned char ctrl, unsigned char value);

    // Use this for melodic only. Percussions have a different mechanism.
    void Patch(unsigned trackno, unsigned ch, unsigned patchno);

    // Note: The value is 14bit and its effect is somewhat undefined.
    void Wheel(unsigned trackno, unsigned ch, unsigned value);
    
    // Controllers work like this:
    void SetVolume(unsigned trackno, unsigned ch, unsigned volume)
    {
        if(volume > 127) volume = 127;
        Control(trackno, ch, 7, volume);
    }
    
    // FIXME: What unit is this "amount" in??
    void AddDelay(unsigned trackno, unsigned amount)
    {
        // amount = amount * deltaticks / tempo;
        tracks[trackno].AddDelay(amount);
    }
    
    unsigned size() const { return midi.size(); }
    unsigned char operator[] (unsigned ind) const { return midi[ind]; }
    const unsigned char* getdata() const { return (const unsigned char* )&midi[0]; }
    
    void Save(const char* filename);
    bool Load(const char* filename);
};

#if 0
class MUSfile
{
public:
    void Init()
    {
    }
    void Done()
    {
        fwrite("MUS\x1A", 1, 4, fp);
        fputc(0, fp);   /* word scorelen - TODO */
        fputc(0, fp);
        fputc(0, fp);   /* word scorestart - TODO */
        fputc(0, fp);
        fputc(15, fp);  /* word count of primary channels */
        fputc(0, fp);
        fputc(0, fp);   /* word count of secondary channels */
        fputc(0, fp);  
    }
};
#endif

#endif
